"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    atom.py -- produce atom feed file of recent torrent updates
    Copyright (C) 2017  Pavel Pletenev <cpp.create@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import getpass
import pwd
import grp
import traceback
import time

import yaml

from ...optconf import Option
from ...optconf.converters import as_string_or_none
from ...optconf.converters import as_path_or_none

from . import BaseConfetti
from . import templated


# =====
def get_uid(user):
    return pwd.getpwnam(user)[2]


def get_gid(group):
    return grp.getgrnam(group)[2]


def get_user_groups(user):
    groups = [
        group.gr_name
        for group in grp.getgrall()
        if user in group.gr_mem
    ]
    gid = pwd.getpwnam(user).pw_gid
    groups.append(grp.getgrgid(gid).gr_name)
    return [grp.getgrnam(group).gr_gid for group in groups]


class UserError(Exception):
    pass


# =====
class Plugin(BaseConfetti):  # pylint: disable=too-many-instance-attributes
    PLUGIN_NAME = "atom"

    def __init__(self,  # pylint: disable=super-init-not-called,too-many-arguments
                 history_path, path, url, user, group, template, html, **kwargs):
        self._init_bases(**kwargs)

        self._history_path = history_path
        self._path = path
        self._url = url
        self._user = (get_uid(user) if user else -1)
        self._group = (get_gid(group) if group else -1)
        if self._group != -1:
            if self._group not in get_user_groups(getpass.getuser()):
                raise UserError(
                    "I wouldn't be able to edit {path} with current user if I chown it for"
                    " uid {uid} and gid {gid}".format(path=path, uid=self._user, gid=self._group),
                )
        self._template_path = template
        self._html = html

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "history_path": Option(default="emonoda_history.yaml", type=as_path_or_none, help="Destination email address"),
            "path":         Option(default="atom.xml", type=as_path_or_none, help="Destination email address"),
            "url":          Option(default="http://localhost/", type=as_string_or_none, help="Feed server url"),
            "user":         Option(default=None, type=as_string_or_none, help="Server user"),
            "group":        Option(default=None, type=as_string_or_none, help="Server user group"),
            "template":     Option(default=None, type=as_path_or_none, help="Mako template file name"),
            "html":         Option(default=True, help="HTML or plaintext feed")
        })

    # ===

    def send_results(self, source, results):
        results_set = []
        try:
            with open(self._history_path) as history_file:
                results_set = yaml.load(history_file)
        except Exception:
            traceback.print_exc()
        results["ctime"] = time.time()
        results_set.insert(0, results)
        results_set = results_set[:20]
        built_in = (self._template_path is None)
        with open(self._path, "w") as atom_file:
            atom_file.write(templated(
                name=("atom.{ctype}.{source}.mako" if built_in else self._template_path).format(
                    ctype=("html" if self._html else "plain"),
                    source=source,
                    results_set=results_set,
                ),
                built_in=built_in,
                source=source,
                results_set=results_set,
                settings=dict(url=self._url)
            ))
        os.chmod(self._path, 0o664)
        os.chown(self._path, self._user, self._group)
        with open(self._history_path, "w") as history_file:
            history_file.write(yaml.dump(results_set))
        del results["ctime"]



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import importlib
import functools
import os


# =====
def get_client_class(name):
    return _get_classes()["clients"][name]


def get_tracker_class(name):
    return _get_classes()["trackers"][name]


def get_confetti_class(name):
    return _get_classes()["confetti"][name]


@functools.lru_cache()
def _get_classes():
    classes = {}
    for sub in ("clients", "trackers", "confetti"):
        classes.setdefault(sub, {})
        sub_path = os.path.join(os.path.dirname(__file__), sub)
        for file_name in os.listdir(sub_path):
            if not file_name.startswith("__") and file_name.endswith(".py"):
                module_name = file_name[:-3]
                module = importlib.import_module("emonoda.plugins.{}.{}".format(sub, module_name))
                plugin_class = getattr(module, "Plugin")
                classes[sub][plugin_class.PLUGIN_NAME] = plugin_class
    return classes


# =====
def _get_bases(mro):
    return tuple(
        cls for cls in mro
        if set(cls.__bases__).intersection((BasePlugin, BaseExtension))
    )


class BasePlugin:
    PLUGIN_NAME = None

    @classmethod
    def get_options(cls):
        return {}

    @classmethod
    def get_bases(cls):
        return _get_bases(cls.__mro__)

    # ===

    def _init_bases(self, **kwargs):
        assert self.PLUGIN_NAME is not None
        for parent in _get_bases(self.__class__.__mro__):
            parent.__init__(self, **kwargs)

    @classmethod
    def _get_merged_options(cls, params=None):
        merged = {}
        for parent in _get_bases(cls.__mro__):
            merged.update(parent.get_options())
        merged.update(params or {})
        return merged


class BaseExtension:
    @classmethod
    def get_options(cls):
        return {}



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from . import nnmclub_to


# =====
class Plugin(nnmclub_to.Plugin):
    PLUGIN_NAME = _NNM_DOMAIN = "ipv6.nnmclub.to"

    _SITE_VERSION = 0
    _SITE_FINGERPRINT_URL = "http://{}".format(_NNM_DOMAIN)

    _COMMENT_REGEXP = re.compile(r"http://ipv6\.(nnm-club\.(me|ru)|nnmclub\.to)/forum/viewtopic\.php\?p=(?P<torrent_id>\d+)")

    _TORRENT_SCRAPE_URL = "http://bt.{}:2710/scrape.php?info_hash={{scrape_hash}}".format(_NNM_DOMAIN)

    _DOWNLOAD_ID_URL = "http://{}/forum/viewtopic.php?p={{torrent_id}}".format(_NNM_DOMAIN)
    _DOWNLOAD_URL = "http://{}//forum/download.php?id={{download_id}}".format(_NNM_DOMAIN)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


# pylint: skip-file
# infinite recursion

import operator

import yaml

from . import Section


# =====
def make_config_dump(config):
    return "\n".join(_inner_make_dump(config))


def _inner_make_dump(config, path=()):
    lines = []
    for (key, value) in sorted(config.items(), key=operator.itemgetter(0)):
        indent = len(path) * "    "
        if isinstance(value, Section):
            lines.append("{}{}:".format(indent, key))
            lines += _inner_make_dump(value, path + (key,))
            lines.append("")
        else:
            default = config._get_default(key)  # pylint: disable=protected-access
            comment = config._get_help(key)  # pylint: disable=protected-access
            if default == value:
                lines.append("{}{}: {} # {}".format(indent, key, _make_yaml(value), comment))
            else:
                lines.append("{}# {}: {} # {}".format(indent, key, _make_yaml(default), comment))
                lines.append("{}{}: {}".format(indent, key, _make_yaml(value)))
    return lines


def _make_yaml(value):
    return yaml.dump(value, allow_unicode=True).replace("\n...\n", "").strip()



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import http.cookiejar
import re

from . import BaseTracker
from . import WithLogin
from . import WithCaptcha
from . import WithCheckHash
from . import WithFetchCustom


# =====
class Plugin(BaseTracker, WithLogin, WithCaptcha, WithCheckHash, WithFetchCustom):
    PLUGIN_NAME = "rutracker.org"

    _SITE_VERSION = 4
    _SITE_ENCODING = "cp1251"
    _SITE_RETRY_CODES = (503, 404)

    _SITE_FINGERPRINT_URL = "https://rutracker.org/forum/index.php"
    _SITE_FINGERPRINT_TEXT = "href=\"//static.t-ru.org/favicon.ico\" type=\"image/x-icon\""

    _COMMENT_REGEXP = re.compile(r"https?://rutracker\.org/forum/viewtopic\.php\?t=(?P<torrent_id>\d+)")

    _TORRENT_HASH_URL = "https://rutracker.org/forum/viewtopic.php?t={torrent_id}"
    _TORRENT_HASH_REGEXP = re.compile(r"<span id=\"tor-hash\">(?P<torrent_hash>[a-fA-F0-9]{40})</span>")

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    # ===

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)
        torrent_id = self._COMMENT_REGEXP.match(torrent.get_comment()).group("torrent_id")
        self._cookie_jar.set_cookie(http.cookiejar.Cookie(
            version=0,
            name="bb_dl",
            value=torrent_id,
            port=None,
            port_specified=False,
            domain="",
            domain_specified=False,
            domain_initial_dot=False,
            path="/forum/",
            path_specified=True,
            secure=True,
            expires=None,
            discard=True,
            comment=None,
            comment_url=None,
            rest={"HttpOnly": None},
            rfc2109=False,
        ))
        data = self._read_url(
            url="https://rutracker.org/forum/dl.php?t={}".format(torrent_id),
            data=b"",
            headers={
                "Referer": "https://rutracker.org/forum/viewtopic.php?t={}".format(torrent_id),
                "Origin":  "https://rutracker.org",
            }
        )
        self._assert_valid_data(data)
        return data

    def login(self):
        self._assert_required_user_passwd()

        post = {
            "login_username": self._encode(self._user),
            "login_password": self._encode(self._passwd),
            "login":          b"\xc2\xf5\xee\xe4",
        }
        page = self._read_login(post)

        cap_static_regexp = re.compile(r"\"//(static\.t-ru\.org/captcha/[^\"]+)\"")
        cap_static_match = cap_static_regexp.search(page)
        if cap_static_match is not None:
            cap_sid_match = re.search(r"name=\"cap_sid\" value=\"([a-zA-Z0-9]+)\"", page)
            cap_code_match = re.search(r"name=\"(cap_code_[a-zA-Z0-9]+)\"", page)
            self._assert_auth(cap_sid_match is not None, "Unknown cap_sid")
            self._assert_auth(cap_code_match is not None, "Unknown cap_code")

            post[cap_code_match.group(1)] = self._captcha_decoder("https://{}".format(cap_static_match.group(1)))
            post["cap_sid"] = cap_sid_match.group(1)
            page = self._read_login(post)
            self._assert_auth(cap_static_regexp.search(page) is None, "Invalid user, password or captcha")

    def _read_login(self, post):
        return self._decode(self._read_url(
            url="https://rutracker.org/forum/login.php",
            data=self._encode(urllib.parse.urlencode(post)),
        ))



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os

import chardet


# =====
def make_sub_name(path, prefix, suffix):
    return os.path.join(
        os.path.dirname(path),
        prefix + os.path.basename(path) + suffix,
    )


def sorted_paths(paths, get=None):
    if get is None:
        # def for speed
        def get_path_nulled(path):
            return path.replace(os.path.sep, "\0")
    else:
        def get_path_nulled(item):
            return item[get].replace(os.path.sep, "\0")
    return sorted(paths, key=get_path_nulled)


def get_decoded_path(path):
    try:
        path.encode()
        return path
    except UnicodeEncodeError:
        path_bytes = os.fsencode(path)
        try:
            return path_bytes.decode("cp1251")
        except UnicodeDecodeError:
            encoding = chardet.detect(path)["encoding"]
            assert encoding is not None, "Can't determine encoding for bytes string: '{}'".format(repr(path_bytes))
            return path_bytes.decode(encoding)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import shlex
import itertools
import argparse
import datetime

from ..plugins.clients import NoSuchTorrentError

from ..helpers import tcollection

from .. import tfile
from .. import fmt
from .. import tools

from . import init
from . import get_configured_log
from . import get_configured_client


# =====
def format_size_pretty(torrent):
    return fmt.format_size(torrent.get_size())


def format_announce(torrent):
    return (torrent.get_announce() or "")


def format_announce_list(torrent):
    return list(itertools.chain.from_iterable(torrent.get_announce_list()))


def format_announce_list_pretty(torrent):
    return " ".join(format_announce_list(torrent))


def format_creation_date(torrent):
    return (torrent.get_creation_date() or "")


def format_creation_date_pretty(torrent):
    date = torrent.get_creation_date()
    return str(datetime.datetime.utcfromtimestamp(date) if date is not None else "")


def format_created_by(torrent):
    return (torrent.get_created_by() or "")


def format_provides(torrent):
    return tools.sorted_paths(torrent.get_files())


def format_is_private(torrent):
    return str(int(torrent.is_private()))


def format_is_private_pretty(torrent):
    return ("yes" if torrent.is_private() else "no")


def format_comment(torrent):
    return (torrent.get_comment() or "")


def _format_client_method(torrent, client, method_name, *args, **kwargs):
    assert client is not None, "Required a client"
    try:
        return getattr(client, method_name)(torrent, *args, **kwargs)
    except NoSuchTorrentError:
        return ""


def format_client_path(torrent, client):
    return _format_client_method(torrent, client, "get_full_path")


def format_client_prefix(torrent, client):
    return _format_client_method(torrent, client, "get_data_prefix")


def format_client_customs(torrent, client, to_show_customs):
    assert client is not None, "Required a client"
    if len(to_show_customs) == 0:
        return ""
    try:
        customs = client.get_customs(torrent, to_show_customs)
    except NoSuchTorrentError:
        return ""
    return " ".join(sorted(
        "{}={}".format(key, shlex.quote(str(value or "")))
        for (key, value) in customs.items()
    ))


def _make_formatted_tree(files, depth=0, prefix="    "):
    text = ""
    for (name, sub) in tools.sorted_paths(files.items(), 0):
        text += prefix + "*   " * depth + name + "\n"
        text += _make_formatted_tree(sub, depth + 1)
    return text


def format_files_tree(torrent):
    tree = {}
    for path in torrent.get_files():
        parts = os.path.normpath(path).split(os.path.sep)
        local = tree
        for (index, part) in enumerate(parts):
            if index != 0 and len(part) == 0:
                continue
            local.setdefault(part, {})
            local = local[part]
    return _make_formatted_tree(tree)


def print_pretty_meta(torrent, client, to_show_customs, log):
    log.print("{blue}Path:{reset}           %s", (torrent.get_path(),))
    log.print("{blue}Name:{reset}           %s", (torrent.get_name(),))
    log.print("{blue}Hash:{reset}           %s", (torrent.get_hash(),))
    log.print("{blue}Size:{reset}           %s", (format_size_pretty(torrent),))
    log.print("{blue}Announce:{reset}       %s", (format_announce(torrent),))
    log.print("{blue}Announce list:{reset}  %s", (format_announce_list_pretty(torrent),))
    log.print("{blue}Creation date:{reset}  %s", (format_creation_date_pretty(torrent),))
    log.print("{blue}Created by:{reset}     %s", (format_created_by(torrent),))
    log.print("{blue}Private:{reset}        %s", (format_is_private_pretty(torrent),))
    log.print("{blue}Comment:{reset}        %s", (format_comment(torrent),))
    if client is not None:
        log.print("{blue}Client path:{reset}    %s", (format_client_path(torrent, client),))
        if len(to_show_customs) != 0:
            log.print("{blue}Client customs:{reset} %s", (format_client_customs(torrent, client, to_show_customs),))
    if torrent.is_single_file():
        log.print("{blue}Provides:{reset}       %s", (tuple(torrent.get_files())[0],))
    else:
        log.print("{blue}Provides:{reset}\n%s", (format_files_tree(torrent),))


def print_value(header, value, without_headers, log):
    if without_headers:
        log.print("%s", str(value))
    else:
        log.print("{blue}%s:{reset} %s", (header, str(value)))


# ===== Main =====
def main():  # pylint: disable=too-many-locals
    options = config = None  # Makes pylint happy
    actions = [
        (option, option[2:].replace("-", "_"), method)
        for (option, method) in (
            ("--path",                 tfile.Torrent.get_path),
            ("--name",                 tfile.Torrent.get_name),
            ("--hash",                 tfile.Torrent.get_hash),
            ("--comment",              format_comment),
            ("--size",                 tfile.Torrent.get_size),
            ("--size-pretty",          format_size_pretty),
            ("--announce",             format_announce),
            ("--announce-list",        format_announce_list),
            ("--announce-list-pretty", format_announce_list_pretty),
            ("--creation-date",        format_creation_date),
            ("--creation-date-pretty", format_creation_date_pretty),
            ("--created-by",           format_created_by),
            ("--provides",             format_provides),
            ("--is-private",           format_is_private),
            ("--is-private-pretty",    format_is_private_pretty),
            ("--client-path",          lambda torrent: format_client_path(torrent, client)),
            ("--client-prefix",        lambda torrent: format_client_prefix(torrent, client)),
            ("--client-customs",       lambda torrent: format_client_customs(torrent, client, config.emfile.show_customs)),
            ("--make-magnet",          lambda torrent: torrent.make_magnet(options.magnet_fields)),
        )
    ]

    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emfile",
        description="Show a metadata of torrent file",
        parents=[parent_parser],
    )
    for (option, dest, _) in actions:
        args_parser.add_argument(option, dest=dest, action="store_true")
    args_parser.add_argument("--without-headers", action="store_true")
    args_parser.add_argument("--magnet-fields", nargs="+", default=(), metavar="<fields>", choices=("names", "trackers", "size"))
    args_parser.add_argument("-v", "--verbose", action="store_true")
    args_parser.add_argument("torrents", nargs="+", metavar="<path>")
    options = args_parser.parse_args(argv[1:])

    to_print = [
        (option[2:], method)
        for (option, dest, method) in actions
        if getattr(options, dest)
    ]
    torrents = tcollection.find(config.core.torrents_dir, options.torrents, False)

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, (not options.verbose), sys.stderr) as log_stderr:
            client = get_configured_client(
                config=config,
                required=False,
                with_customs=bool(config.emfile.show_customs),
                log=log_stderr,
            )

            for torrent in torrents:
                if len(to_print) == 0:
                    print_pretty_meta(torrent, client, config.emfile.show_customs, log_stdout)
                else:
                    for (header, method) in to_print:
                        retval = method(torrent)
                        if isinstance(retval, (list, tuple)):
                            for item in retval:
                                print_value(header, item, options.without_headers, log_stdout)
                        else:
                            print_value(header, retval, options.without_headers, log_stdout)
                if len(torrents) > 1:
                    log_stdout.print()


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import socket
import xmlrpc.client
import operator
import math
import time
import argparse


# =====
def get_summary(server, hashes):
    mapping = (
        ("get_peers_accounted",  "leechers"),
        ("is_hash_checking",     "is_checking"),
        ("get_completed_chunks", "completed_chunks"),
        ("get_chunks_hashed",    "hashed_chunks"),
        ("get_size_chunks",      "size_chunks"),
        ("get_message",          "msg"),
    )
    multicall = xmlrpc.client.MultiCall(server)
    for torrent_hash in hashes:
        for (method_name, _) in mapping:
            getattr(multicall.d, method_name)(torrent_hash)
    rows = tuple(multicall())
    rows = tuple(
        dict(zip(map(operator.itemgetter(1), mapping), rows[count:count + len(mapping)]))
        for count in range(0, len(rows), len(mapping))
    )

    summary = dict.fromkeys(("total", "dn", "up", "errors"), 0)
    for row in rows:
        if row["leechers"]:
            summary["up"] += 1
        chunks_processing = (row["completed_chunks"] if row["is_checking"] else row["hashed_chunks"])
        done = math.floor(chunks_processing / row["size_chunks"] * 1000)
        if done != 1000:
            summary["dn"] += 1
        if len(row["msg"]) and row["msg"] != "Tracker: [Tried all trackers.]":
            summary["errors"] += 1
    summary["total"] = len(hashes)
    return summary


def print_stat(client_url, host, interval, with_dht, with_summary):
    server = xmlrpc.client.ServerProxy(client_url)
    while True:
        multicall = xmlrpc.client.MultiCall(server)
        multicall.get_down_rate()  # Download rate
        multicall.get_download_rate()  # Download rate limit
        multicall.get_down_total()  # Downloaded
        multicall.get_up_rate()  # Upload rate
        multicall.get_upload_rate()  # Upload rate limit
        multicall.get_up_total()  # Uploaded
        multicall.dht_statistics()
        multicall.download_list()
        values = tuple(multicall())

        metrics = [
            ("gauge-dn_rate",       values[0]),
            ("gauge-dn_rate_limit", values[1]),
            ("bytes-dn_total",      values[2]),
            ("gauge-up_rate",       values[3]),
            ("gauge-up_rate_limit", values[4]),
            ("bytes-up_total",      values[5]),
        ]
        if with_dht:
            metrics += [
                ("gauge-dht_active",           values[6]["active"]),
                ("count-dht_nodes",            values[6]["nodes"]),
                ("count-dht_cycle",            values[6]["cycle"]),
                ("count-dht_torrents",         values[6]["torrents"]),
                ("count-dht_buckets",          values[6]["buckets"]),
                ("count-dht_replies_received", values[6]["replies_received"]),
                ("count-dht_peers",            values[6]["peers"]),
                ("count-dht_peers_max",        values[6]["peers_max"]),
                ("count-dht_errors_caught",    values[6]["errors_caught"]),
                ("count-dht_errors_received",  values[6]["errors_received"]),
                ("count-dht_queries_sent",     values[6]["queries_sent"]),
                ("count-dht_queries_received", values[6]["queries_received"]),
                ("bytes-dht_bytes_written",    values[6]["bytes_written"]),
                ("bytes-dht_bytes_read",       values[6]["bytes_read"]),
            ]
        if with_summary:
            summary = get_summary(server, values[7])
            metrics += [
                ("count-summary_total",  summary["total"]),
                ("count-summary_dn",     summary["dn"]),
                ("count-summary_up",     summary["up"]),
                ("count-summary_errors", summary["errors"]),
            ]

        for (key, value) in metrics:
            print("PUTVAL {}/rtorrent/{} interval={} N:{}".format(host, key, interval, value), flush=True)
        time.sleep(interval)


# ===== Main =====
def main():
    args_parser = argparse.ArgumentParser(description="Prints collectd stat in plaintext protocol")
    args_parser.add_argument("--with-dht", action="store_true")
    args_parser.add_argument("--with-summary", action="store_true")
    args_parser.add_argument("-n", "--host", default=os.getenv("COLLECTD_HOSTNAME", "localhost"), metavar="<hostname>")
    args_parser.add_argument("-i", "--interval", default=os.getenv("COLLECTD_INTERVAL", 60), type=float, metavar="<seconds>")
    args_parser.add_argument("-t", "--timeout", default=5.0, type=float, metavar="<seconds>")
    args_parser.add_argument("--client-url", default="http://localhost/RPC2", metavar="<url>")

    options = args_parser.parse_args(sys.argv[1:])
    socket.setdefaulttimeout(options.timeout)
    try:
        print_stat(
            client_url=options.client_url,
            host=options.host,
            interval=options.interval,
            with_dht=options.with_dht,
            with_summary=options.with_summary,
        )
    except (SystemExit, KeyboardInterrupt):
        pass


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import argparse

from ..plugins.clients import NoSuchTorrentError

from ..helpers import tcollection

from .. import tfile

from . import init
from . import get_configured_log
from . import get_configured_client


# ===== Main =====
def main():
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emrm",
        description="Remove a torrent from client",
        parents=[parent_parser],
    )
    args_parser.add_argument("-v", "--verbose", action="store_true")
    args_parser.add_argument("torrents", type=str, nargs="+", metavar="<path/hash>")
    options = args_parser.parse_args(argv[1:])

    with get_configured_log(config, (not options.verbose), sys.stdout) as log_stdout:
        with get_configured_log(config, (not options.verbose), sys.stderr) as log_stderr:
            client = get_configured_client(
                config=config,
                required=True,
                with_customs=False,
                log=log_stderr,
            )

            hashes = []
            for item in tcollection.find(config.core.torrents_dir, options.torrents, True):
                torrent_hash = (item.get_hash() if isinstance(item, tfile.Torrent) else item)
                try:
                    hashes.append((torrent_hash, client.get_file_name(torrent_hash)))
                except NoSuchTorrentError:
                    log_stderr.error("No such torrent: {yellow}%s{reset}", (torrent_hash,))

            if len(hashes) != 0:
                log_stderr.info("Removed:")
                for (torrent_hash, name) in hashes:
                    client.remove_torrent(torrent_hash)
                    log_stdout.print("{yellow}%s{reset} -- %s", (torrent_hash, name))


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Description: bencodode/decode library
    Upstream: https://github.com/flying-sheep/bcode
    License: MIT

    Copyright (c) 2013 Phil Schaf

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# flake8: noqa
# pylint: disable=invalid-name


from io import BytesIO, SEEK_CUR
from string import digits
try: #py 3.3
    from collections.abc import Iterable, Mapping
except ImportError:
    from collections     import Iterable, Mapping

try:
    str = unicode  # pylint: disable=redefined-builtin,undefined-variable
except NameError:
    pass

_TYPE_INT  = b'i'
_TYPE_LIST = b'l'
_TYPE_DICT = b'd'
_TYPE_END  = b'e'
_TYPE_SEP  = b':'
_TYPES_STR = [d.encode() for d in digits]

def assert_btype(byte, typ):
    if byte != typ:
        raise TypeError(
            'Tried to decode type {!r} with identifier {!r} but got identifier {!r} instead'
            .format(TYPES[typ] or 'End', typ, byte))

################
### Decoding ###
################

def _readuntil(f, end=_TYPE_END):
    """Helper function to read bytes until a certain end byte is hit"""
    buf = bytearray()
    byte = f.read(1)
    while byte != end:
        if byte == b'':
            raise ValueError('File ended unexpectedly. Expected end byte {}.'.format(end))
        buf += byte
        byte = f.read(1)
    return buf

def _decode_int(f):
    """
    Integer types are normal ascii integers
    Delimited at the start with 'i' and the end with 'e'
    """
    assert_btype(f.read(1), _TYPE_INT)
    return int(_readuntil(f))

def _decode_buffer(f):
    """
    String types are normal (byte)strings
    starting with an integer followed by ':'
    which designates the string’s length.

    Since there’s no way to specify the byte type
    in bencoded files, we have to guess
    """
    strlen = int(_readuntil(f, _TYPE_SEP))
    buf = f.read(strlen)
    if len(buf) != strlen:
        raise ValueError(
            'string expected to be {} bytes long but the file ended after {} bytes'
            .format(strlen, len(buf)))
    try:
        return buf.decode()
    except UnicodeDecodeError:
        return buf

def _decode_list(f):
    assert_btype(f.read(1), _TYPE_LIST)
    ret = []
    item = bdecode(f)
    while item is not None:
        ret.append(item)
        item = bdecode(f)
    return ret

def _decode_dict(f):
    assert_btype(f.read(1), _TYPE_DICT)
    ret = {}
    key = bdecode(f)
    while key is not None:
        assert isinstance(key, (str, bytes))
        ret[key] = bdecode(f)
        key = bdecode(f)
    return ret

TYPES = {
    _TYPE_INT:  _decode_int,
    _TYPE_LIST: _decode_list,
    _TYPE_DICT: _decode_dict,
    _TYPE_END:  None,
    #_TYPE_SEP only appears in strings, not here
}
TYPES.update({byte: _decode_buffer for byte in _TYPES_STR}) #b'0': _decode_buffer, b'1': _decode_buffer, …

def bdecode(f_or_data):
    """
    bdecodes data by looking up the type byte,
    and using it to look up the respective decoding function,
    which in turn is used to return the decoded object

    The parameter can be a file opened in bytes mode,
    bytes or a string (the last of which will be decoded)
    """
    if isinstance(f_or_data, str):
        f_or_data = f_or_data.encode()
    if isinstance(f_or_data, bytes):
        f_or_data = BytesIO(f_or_data)

    #TODO: the following line is the only one that needs readahead.
    #peek returns a arbitrary amount of bytes, so we have to slice.
    if f_or_data.seekable():
        first_byte = f_or_data.read(1)
        f_or_data.seek(-1, SEEK_CUR)
    else:
        #FIXME: muted bug!
        first_byte = f_or_data.peek(1)[:1]  # pylint: disable=no-member
    btype = TYPES.get(first_byte)
    if btype is not None:
        return btype(f_or_data)
    else: #Used in dicts and lists to designate an end
        assert_btype(f_or_data.read(1), _TYPE_END)
        return None

################
### Encoding ###
################

def _encode_int(integer, f):
    f.write(_TYPE_INT)
    f.write(str(integer).encode())
    f.write(_TYPE_END)

def _encode_buffer(string, f):
    """Writes the bencoded form of the input string or bytes"""
    if isinstance(string, str):
        string = string.encode()
    f.write(str(len(string)).encode())
    f.write(_TYPE_SEP)
    f.write(string)

def _encode_iterable(iterable, f):
    f.write(_TYPE_LIST)
    for item in iterable:
        bencode(item, f)
    f.write(_TYPE_END)

def _encode_mapping(mapping, f):
    """Encodes the mapping items in lexical order (spec)"""
    f.write(_TYPE_DICT)
    for key, value in sorted(mapping.items()):
        _encode_buffer(key, f)
        bencode(value, f)
    f.write(_TYPE_END)

def _bencode_to_file(data, f):
    if isinstance(data, int):
        _encode_int(data, f)
    elif isinstance(data, (str, bytes)):
        _encode_buffer(data, f)
    elif isinstance(data, Mapping):
        _encode_mapping(data, f)
    elif isinstance(data, Iterable):
        _encode_iterable(data, f)
    else:
        raise TypeError(
            'the passed value {} of type {} is not bencodable.'
            .format(data, type(data).__name__))

def bencode(data, f=None):
    """
    Writes a serializable data piece to f
    The order of tests is nonarbitrary,
    as strings and mappings are iterable.

    If f is None, it writes to a byte buffer
    and returns a bytestring
    """
    if f is None:
        f = BytesIO()
        _bencode_to_file(data, f)
        return f.getvalue()
    else:
        _bencode_to_file(data, f)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os

from ... import tfile

from .. import BasePlugin
from .. import BaseExtension


# =====
class NoSuchTorrentError(Exception):
    pass


# =====
def build_files(prefix, flist):
    files = {}
    for (path, size) in flist:
        path_list = path.split(os.path.sep)
        name = None
        for index in range(len(path_list)):
            name = os.path.join(prefix, os.path.sep.join(path_list[0:index + 1]))
            files[name] = None
        assert name is not None
        files[name] = {"size": size}
    return files


def hash_or_torrent(method):
    def wrap(self, torrent, *args, **kwargs):
        torrent_hash = (torrent.get_hash() if isinstance(torrent, tfile.Torrent) else torrent)
        return method(self, torrent_hash, *args, **kwargs)
    return wrap


def check_torrent_accessible(method):
    def wrap(self, torrent, prefix=None):
        path = torrent.get_path()
        assert path is not None, "Required Torrent() with local file"
        open(path, "rb").close()  # Check accessible file
        if prefix is not None:
            os.listdir(prefix)  # Check accessible prefix
        return method(self, torrent, prefix)
    return wrap


# =====
class BaseClient(BasePlugin):
    def __init__(self, **_):
        pass

    @hash_or_torrent
    def remove_torrent(self, torrent_hash):
        raise NotImplementedError

    @check_torrent_accessible
    def load_torrent(self, torrent, prefix=None):
        raise NotImplementedError

    def get_hashes(self):
        raise NotImplementedError

    @hash_or_torrent
    def has_torrent(self, torrent_hash):
        raise NotImplementedError

    @hash_or_torrent
    def get_torrent_path(self, torrent_hash):
        raise NotImplementedError

    @hash_or_torrent
    def get_data_prefix(self, torrent_hash):
        raise NotImplementedError

    def get_data_prefix_default(self):
        raise NotImplementedError

    # ===

    @hash_or_torrent
    def get_full_path(self, torrent_hash):
        raise NotImplementedError

    @hash_or_torrent
    def get_file_name(self, torrent_hash):
        raise NotImplementedError

    @hash_or_torrent
    def is_single_file(self, torrent_hash):
        raise NotImplementedError

    @hash_or_torrent
    def get_files(self, torrent_hash, on_fs=False):
        raise NotImplementedError


class WithCustoms(BaseExtension):
    def __init__(self, **_):
        pass

    @classmethod
    def get_custom_keys(cls):
        raise NotImplementedError

    @hash_or_torrent
    def set_customs(self, torrent_hash, customs):  # pylint: disable=unused-argument
        raise NotImplementedError

    @hash_or_torrent
    def get_customs(self, torrent_hash, keys):  # pylint: disable=unused-argument
        raise NotImplementedError



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from ...optconf import Option

from . import BaseTracker
from . import WithLogin
from . import WithCheckScrape
from . import WithFetchByDownloadId


# =====
class Plugin(BaseTracker, WithLogin, WithCheckScrape, WithFetchByDownloadId):
    PLUGIN_NAME = _NNM_DOMAIN = "nnmclub.to"

    _SITE_VERSION = 2
    _SITE_ENCODING = "cp1251"

    _SITE_FINGERPRINT_URL = "https://{}".format(_NNM_DOMAIN)
    _SITE_FINGERPRINT_TEXT = "<link rel=\"canonical\" href=\"http://nnmclub.to/\">"

    _COMMENT_REGEXP = re.compile(r"https?://(nnm-club\.(me|ru)|nnmclub\.to)/forum/viewtopic\.php\?p=(?P<torrent_id>\d+)")

    _TORRENT_SCRAPE_URL = "http://bt.{}:2710/scrape.php?info_hash={{scrape_hash}}".format(_NNM_DOMAIN)

    _DOWNLOAD_ID_URL = "https://{}/forum/viewtopic.php?p={{torrent_id}}".format(_NNM_DOMAIN)
    _DOWNLOAD_ID_REGEXP = re.compile(r"filelst.php\?attach_id=(?P<download_id>[a-zA-Z0-9]+)")
    _DOWNLOAD_URL = "https://{}//forum/download.php?id={{download_id}}".format(_NNM_DOMAIN)

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "timeout": Option(default=20.0, help="Timeout for HTTP client"),
        })

    def login(self):
        self._login_using_post(
            url="https://{}/forum/login.php".format(self._NNM_DOMAIN),
            post={
                "username": self._encode(self._user),
                "password": self._encode(self._passwd),
                "redirect": b"",
                "login":    b"\xc2\xf5\xee\xe4",
            },
            ok_text="class=\"mainmenu\">Выход [ {} ]</a>".format(self._user),
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse

from ...optconf import Option
from ...optconf.converters import as_string_list

from ... import web

from . import BaseConfetti
from . import WithProxy


# =====
class Plugin(BaseConfetti, WithProxy):
    PLUGIN_NAME = "pushover"

    def __init__(self, user_key, api_key, devices, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)

        self._user_key = user_key
        self._api_key = api_key
        self._devices = devices

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "user_key": Option(default="CHANGE_ME", type=str, help="User key"),
            "api_key":  Option(default="CHANGE_ME", type=str, help="API/Application key"),
            "devices":  Option(default=[], type=as_string_list, help="Devices list (empty for all)"),
        })

    # ===

    def send_results(self, source, results):
        for result in results["affected"].values():
            self._notify(
                title="Emonoda ({})".format(source),
                message=result["torrent"].get_name(),
            )

    # ===

    def _notify(self, title, message):
        # https://pushover.net/api
        post = {
            "token":   self._api_key,
            "user":    self._user_key,
            "title":   title,
            "message": message,
        }
        web.read_url(
            opener=web.build_opener(proxy_url=self._proxy_url),
            url="https://api.pushover.net/1/messages.json",
            data=urllib.parse.urlencode(post).encode("utf-8"),
            timeout=self._timeout,
            retries=self._retries,
            retries_sleep=self._retries_sleep,
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import base64

from ...optconf import Option
from ...optconf import SecretOption
from ...optconf.converters import as_string_or_none

from . import BaseClient
from . import NoSuchTorrentError
from . import hash_or_torrent
from . import check_torrent_accessible
from . import build_files

try:
    import transmissionrpc
except ImportError:
    transmissionrpc = None


# =====
class Plugin(BaseClient):
    # API description:
    #   http://pythonhosted.org/transmissionrpc/
    #   https://trac.transmissionbt.com/browser/trunk/extras/rpc-spec.txt

    PLUGIN_NAME = "transmission"

    def __init__(self, url, user, passwd, timeout, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)

        if transmissionrpc is None:
            raise RuntimeError("Required module transmissionrpc")

        self._client = transmissionrpc.Client(
            address=url,
            user=user,
            password=passwd,
            timeout=timeout,
        )

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "url":     Option(default="http://localhost:9091/transmission/rpc", help="Transmission HTTP-RPC URL"),
            "timeout": Option(default=10.0, help="Timeout for HTTP-RPC"),
            "user":    Option(default=None, type=as_string_or_none, help="HTTP login"),
            "passwd":  SecretOption(default=None, type=as_string_or_none, help="HTTP password"),
        })

    # ===

    @hash_or_torrent
    def remove_torrent(self, torrent_hash):
        self._get_torrent_obj(torrent_hash)  # XXX: raise NoSuchTorrentError if torrent does not exist
        self._client.remove_torrent(torrent_hash)

    @check_torrent_accessible
    def load_torrent(self, torrent, prefix=None):
        kwargs = {"paused": False}
        if prefix is not None:
            kwargs["download_dir"] = prefix
        self._client.add_torrent(base64.b64encode(torrent.get_data()).decode("utf-8"), **kwargs)

    @hash_or_torrent
    def has_torrent(self, torrent_hash):
        try:
            self._get_torrent_obj(torrent_hash)
            return True
        except NoSuchTorrentError:
            return False

    def get_hashes(self):
        return [
            str(item.hashString.lower())
            for item in self._client.get_torrents(arguments=("id", "hashString"))
        ]

    @hash_or_torrent
    def get_torrent_path(self, torrent_hash):
        return self._get_torrent_prop(torrent_hash, "torrentFile")

    @hash_or_torrent
    def get_data_prefix(self, torrent_hash):
        return self._get_torrent_prop(torrent_hash, "downloadDir")

    def get_data_prefix_default(self):
        session = self._client.get_session()
        assert session is not None
        return session.download_dir

    # ===

    @hash_or_torrent
    def get_full_path(self, torrent_hash):
        torrent_obj = self._get_torrent_obj(torrent_hash, ("name", "downloadDir"))
        return os.path.join(torrent_obj.downloadDir, torrent_obj.name)

    @hash_or_torrent
    def get_file_name(self, torrent_hash):
        return self._get_torrent_prop(torrent_hash, "name")

    @hash_or_torrent
    def is_single_file(self, torrent_hash):
        files = self._get_files(torrent_hash)
        if len(files) > 1:
            return False
        return (os.path.sep not in list(files.values())[0]["name"])

    @hash_or_torrent
    def get_files(self, torrent_hash, on_fs=False):
        prefix = (self.get_data_prefix(torrent_hash) if on_fs else "")
        flist = [
            (item["name"], item["size"])
            for item in self._get_files(torrent_hash).values()
        ]
        return build_files(prefix, flist)

    # ===

    def _get_torrent_prop(self, torrent_hash, prop):
        return getattr(self._get_torrent_obj(torrent_hash, (prop,)), prop)

    def _get_torrent_obj(self, torrent_hash, props=()):
        props = set(props).union(("id", "hashString"))
        try:
            torrent_obj = self._client.get_torrent(torrent_hash, arguments=tuple(props))
        except KeyError as err:
            if str(err) == "\'Torrent not found in result\'":
                raise NoSuchTorrentError("Unknown torrent hash")
            raise
        assert str(torrent_obj.hashString).lower() == torrent_hash
        return torrent_obj

    def _get_files(self, torrent_hash):
        files = self._client.get_files(torrent_hash)
        if len(files) == 0:
            raise NoSuchTorrentError("Unknown torrent hash")
        assert len(files) == 1
        files = list(files.values())[0]
        assert len(files) > 0
        return files



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from datetime import datetime

from . import BaseTracker
from . import WithLogin
from . import WithCheckTime
from . import WithFetchByTorrentId


# =====
class Plugin(BaseTracker, WithLogin, WithCheckTime, WithFetchByTorrentId):
    PLUGIN_NAME = "kinozal.tv"

    _SITE_VERSION = 0
    _SITE_ENCODING = "cp1251"

    _SITE_FINGERPRINT_URL = "http://kinozal.tv/"
    _SITE_FINGERPRINT_TEXT = "<title>Торрент трекер Кинозал.ТВ</title>"

    _COMMENT_REGEXP = re.compile(r"http://kinozal\.tv/details\.php\?id=(?P<torrent_id>\d+)")

    _TIMEZONE_URL = "http://kinozal.tv/my.php"
    _TIMEZONE_REGEXP = re.compile(r"<option value=\"(?P<delta>-?\d{1,3})\" selected")
    _TIMEZONE_PREFIX = "Etc/GMT"

    _DOWNLOAD_URL = "http://dl.kinozal.tv/download.php?id={torrent_id}"
    _DOWNLOAD_PAYLOAD = b""

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    def init_tzinfo(self):
        timezone = None
        page = self._decode(self._read_url(self._TIMEZONE_URL))
        delta_match = self._TIMEZONE_REGEXP.search(page)
        if delta_match:
            timezone_gmt_number = (int(delta_match.group("delta")) + 180) // 60  # Moscow Timezone GMT+3 -> 3 * 60 = 180
            if timezone_gmt_number > 0:
                timezone = self._TIMEZONE_PREFIX + "+" + str(timezone_gmt_number)
            else:
                timezone = self._TIMEZONE_PREFIX + str(timezone_gmt_number)
        self._tzinfo = self._select_tzinfo(timezone)

    def fetch_time(self, torrent):
        self._assert_match(torrent)
        page = self._decode(self._read_url(torrent.get_comment()))

        for label in ("Обновлен", "Залит"):
            date_match = re.search(r"<li>%s<span class=\"floatright green n\">"
                                   r"(\d{1,2}) ([А-Яа-я]{3,8}) (\d{4}) в (\d{2}:\d{2})"
                                   r"</span></li>" % (label), page)
            if date_match is not None:
                break
        self._assert_logic(date_match is not None, "Upload date not found")

        months = {
            "января":   "01",
            "февраля":  "02",
            "марта":    "03",
            "апреля":   "04",
            "мая":      "05",
            "июня":     "06",
            "июля":     "07",
            "августа":  "08",
            "сентября": "09",
            "октября":  "10",
            "ноября":   "11",
            "декабря":  "12",
        }
        date_str = "%s %s %s %s %s" % (
            date_match.group(1),  # Day
            months[date_match.group(2)],  # Month
            date_match.group(3),  # Year
            date_match.group(4),  # Time
            datetime.now(self._tzinfo).strftime("%z")  # Timezone offset
        )
        upload_time = int(datetime.strptime(date_str, "%d %m %Y %H:%M %z").strftime("%s"))
        return upload_time

    def login(self):
        self._login_using_post(
            url="http://kinozal.tv/takelogin.php",
            post={
                "username": self._encode(self._user),
                "password": self._encode(self._passwd),
                "returnto": "",
            },
            ok_text="href=\"/my.php\"",
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.request
import gzip


# =====
class GzipHandler(urllib.request.BaseHandler):
    def http_request(self, request):
        request.add_header("Accept-Encoding", "gzip")
        return request

    def http_response(self, request, response):  # pylint: disable=unused-argument
        if response.headers.get("Content-Encoding") == "gzip":
            gzip_file = gzip.GzipFile(fileobj=response, mode="r")
            old = response
            response = urllib.request.addinfourl(gzip_file, old.headers, old.url, old.code)
            response.msg = old.msg
        return response

    https_request = http_request
    https_response = http_response



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2016  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys

from ..helpers import surprise

from . import init
from . import get_configured_log
from . import get_configured_confetti


# =====
class _FakeTorrent:
    def get_comment(self):
        return "Test, test, test"


class _FakeTracker:
    PLUGIN_NAME = "example.org"


# ===== Main =====
def main():
    (parent_parser, argv, config) = init()  # pylint: disable=unused-variable
    with get_configured_log(config, False, sys.stderr) as log_stderr:
        results = {
            "affected": {
                "test.torrent": {
                    "diff": {
                        "added":         ("nya.mkv", "nya.srt"),
                        "removed":       ("nyaa.srt", "nyaa.mkv"),
                        "modified":      (),
                        "type_modified": ("list.lst",),
                    },
                    "torrent": _FakeTorrent(),
                    "tracker": _FakeTracker(),
                },
                "test1.torrent": {
                    "diff": {
                        "added":         ("nya.mkv", "nya.srt"),
                        "removed":       ("nyaa.srt", "nyaa.mkv"),
                        "modified":      (),
                        "type_modified": ("list.lst",),
                    },
                    "torrent": _FakeTorrent(),
                    "tracker": _FakeTracker(),
                }
            },
            "passed":          {},
            "not_in_client":   {},
            "unknown":         {},
            "invalid":         {},
            "tracker_error":   {},
            "unhandled_error": {},
        }
        confetti = get_configured_confetti(
            config=config,
            log=log_stderr,
        )
        surprise.deploy_surprise(
            source="emupdate",
            results=results,
            confetti=confetti,
            log=log_stderr,
        )


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import argparse

from ..helpers import tcollection

from .. import fmt

from . import init
from . import get_configured_log
from . import get_configured_client


# =====
def make_path(path, tail_mode):
    try:
        os.makedirs(path)
        if tail_mode is not None:
            os.chmod(path, tail_mode)
    except FileExistsError:
        pass


def link_data(torrent, data_dir_path, link_to_path, mkdir_mode):
    mkdir_path = link_to_path = os.path.abspath(link_to_path)
    if torrent.is_single_file():
        link_to_path = os.path.join(link_to_path, torrent.get_name())
    else:
        mkdir_path = os.path.dirname(link_to_path)

    if os.path.exists(link_to_path):
        raise RuntimeError("{}: link target already exists".format(link_to_path))

    make_path(mkdir_path, mkdir_mode)
    os.symlink(os.path.join(data_dir_path, torrent.get_name(surrogate_escape=True)), link_to_path)


def load_torrents(torrents, client, data_root_path, link_to_path, mkdir_mode, customs):
    for torrent in torrents:
        if client.has_torrent(torrent):
            raise RuntimeError("{}: already loaded".format(torrent.get_path()))

    if data_root_path is None:
        data_root_path = client.get_data_prefix_default()

    for torrent in torrents:
        dir_name = os.path.basename(torrent.get_path()) + ".data"
        data_dir_path = os.path.join(data_root_path, dir_name[0], dir_name)
        make_path(data_dir_path, mkdir_mode)

        if link_to_path is not None:
            link_data(torrent, data_dir_path, link_to_path, mkdir_mode)

        client.load_torrent(torrent, data_dir_path)
        if len(customs) != 0:
            client.set_customs(torrent, {
                key: fmt.format_now(value)
                for (key, value) in customs.items()
            })


def parse_customs(items):
    customs = {}
    for item in filter(None, map(str.strip, items)):
        (key, value) = map(str.strip, (item.split("=", 1) + [""])[:2])
        customs[key] = value
    return customs


# ===== Main =====
def main():
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emload",
        description="Load torrent to client",
        parents=[parent_parser],
    )
    args_parser.add_argument("-l", "--link-to", default=None, metavar="<path>")
    args_parser.add_argument("--set-customs", default=[], nargs="+", metavar="<key=value>")
    args_parser.add_argument("-v", "--verbose", action="store_true")
    args_parser.add_argument("torrents", nargs="+", metavar="<path>")
    options = args_parser.parse_args(argv[1:])

    if len(options.torrents) > 1 and options.link_to is not None:
        raise RuntimeError("Option -l/--link-to be used with only one torrent")

    customs = dict(config.emload.set_customs)
    customs.update(parse_customs(options.set_customs))

    torrents = tcollection.find(config.core.torrents_dir, options.torrents, False)

    with get_configured_log(config, (not options.verbose), sys.stderr) as log_stderr:
        client = get_configured_client(
            config=config,
            required=True,
            with_customs=bool(customs),
            log=log_stderr,
        )

        load_torrents(
            torrents=torrents,
            client=client,
            data_root_path=config.core.data_root_dir,
            link_to_path=options.link_to,
            mkdir_mode=config.emload.mkdir_mode,
            customs=customs,
        )


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import socket
import smtplib
import email.mime.multipart
import email.mime.text
import email.header
import email.utils
import contextlib
import time

from ...optconf import Option
from ...optconf.converters import as_string_or_none
from ...optconf.converters import as_string_list
from ...optconf.converters import as_path_or_none

from . import BaseConfetti
from . import templated


# =====
class Plugin(BaseConfetti):  # pylint: disable=too-many-instance-attributes
    PLUGIN_NAME = "email"

    def __init__(self,  # pylint: disable=super-init-not-called,too-many-arguments
                 to, cc, subject, sender, html, server, port, ssl, user, passwd, template, **kwargs):
        self._init_bases(**kwargs)

        self._to = to
        self._cc = cc
        self._subject = subject
        self._sender = sender
        self._html = html
        self._template_path = template

        self._server = server
        self._port = port
        self._ssl = ssl
        self._user = user
        self._passwd = passwd

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "to":       Option(default=["root@localhost"], type=as_string_list, help="Destination email address"),
            "cc":       Option(default=[], type=as_string_list, help="Email 'CC' field"),
            "subject":  Option(default="{source} report: you have {affected} new torrents ^_^", help="Email subject"),
            "sender":   Option(default="root@localhost", help="Email 'From' field"),
            "html":     Option(default=True, help="HTML or plaintext email body"),
            "template": Option(default=None, type=as_path_or_none, help="Mako template file name"),

            "server":   Option(default="localhost", help="Hostname of SMTP server"),
            "port":     Option(default=0, help="Port of SMTP server"),
            "ssl":      Option(default=False, help="Use SMTPS"),
            "user":     Option(default=None, type=as_string_or_none, help="Account on SMTP server"),
            "passwd":   Option(default=None, type=as_string_or_none, help="Passwd for account on SMTP server"),
        })

    # ===

    def send_results(self, source, results):
        msg = self._format_message(source, results)
        retries = self._retries
        while True:
            try:
                self._send_message(msg)
                break
            except (
                smtplib.SMTPServerDisconnected,
                smtplib.SMTPConnectError,
                smtplib.SMTPHeloError,
                socket.timeout,
            ):
                if retries == 0:
                    raise
                time.sleep(self._retries_sleep)
                retries -= 1

    # ===

    def _format_message(self, source, results):
        subject_placeholders = {field: len(items) for (field, items) in results.items()}
        subject_placeholders["source"] = source
        built_in = (self._template_path is None)
        return self._make_message(
            subject=self._subject.format(**subject_placeholders),
            body=templated(
                name=("email.{ctype}.{source}.mako" if built_in else self._template_path).format(
                    ctype=("html" if self._html else "plain"),
                    source=source,
                ),
                built_in=built_in,
                source=source,
                results=results,
            ),
        )

    def _make_message(self, subject, body):
        email_headers = {
            "From":    self._sender,
            "To":      ", ".join(self._to),
            "Date":    email.utils.formatdate(localtime=True),
            "Subject": email.header.Header(subject, "utf-8"),
        }
        if len(self._cc) > 0:
            email_headers["CC"] = ", ".join(self._cc)

        msg = email.mime.multipart.MIMEMultipart()
        for (key, value) in email_headers.items():
            msg[key] = value

        msg.attach(email.mime.text.MIMEText(
            _text=body.encode("utf-8"),
            _subtype=("html" if self._html else "plain"),
            _charset="utf-8",
        ))
        return msg

    def _send_message(self, msg):
        smtp_class = (smtplib.SMTP_SSL if self._ssl else smtplib.SMTP)
        with contextlib.closing(smtp_class(
            host=self._server,
            port=self._port,
            timeout=self._timeout,
        )) as client:
            if self._user is not None:
                client.login(self._user, self._passwd)  # pylint: disable=no-member
            client.send_message(msg)  # pylint: disable=no-member



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from . import BaseTracker
from . import WithLogin
from . import WithCheckHash
from . import WithFetchByDownloadId


# =====
class Plugin(BaseTracker, WithLogin, WithCheckHash, WithFetchByDownloadId):
    PLUGIN_NAME = "tfile.co"

    _SITE_VERSION = 4
    _SITE_ENCODING = "cp1251"

    _SITE_FINGERPRINT_URL = "http://tfile.co"
    _SITE_FINGERPRINT_TEXT = "href=\"http://tfile.co/opensearch.xml\""

    _COMMENT_REGEXP = re.compile(r"http://tfile\.(me|ru|co)/forum/viewtopic\.php\?p=(?P<torrent_id>\d+)")

    _TORRENT_HASH_URL = "http://tfile.co/forum/viewtopic.php?p={torrent_id}"
    _TORRENT_HASH_REGEXP = re.compile(r"<td style=\"color:darkgreen\">Info hash:</td>"
                                      r"<td><strong>(?P<torrent_hash>[a-fA-F0-9]{40})</strong></td>")

    _DOWNLOAD_ID_URL = "http://tfile.co/forum/viewtopic.php?p={torrent_id}"
    _DOWNLOAD_ID_REGEXP = re.compile(r"<a href=\"download.php\?id=(?P<download_id>\d+)\""
                                     r" style=\"background:url\(/blueGene/images/topic\.jpg\)")
    _DOWNLOAD_URL = "http://tfile.co/forum/download.php?id={download_id}"

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    def login(self):
        self._login_using_post(
            url="http://tfile.co/login/",
            post={
                "username":  self._encode(self._user),
                "password":  self._encode(self._passwd),
                "autologin": b"",
                "login":     b"",
            },
            ok_text="class=\"nick u\">{}</a>".format(self._user),
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import traceback


# =====
def deploy_surprise(source, results, confetti, log):
    ok = True
    for sender in confetti:
        log.info("Sending {blue}%s{reset} ...", (sender.PLUGIN_NAME,), one_line=True)
        try:
            sender.send_results(source, results)
            log.info("Result {blue}%s{reset} was {green}sent{reset}", (sender.PLUGIN_NAME,))
        except Exception as err:
            log.error("Can't send {red}%s{reset}: {red}%s{reset}(%s)", (sender.PLUGIN_NAME, type(err).__name__, err))
            log.print("%s", ("\n".join("\t" + row for row in traceback.format_exc().strip().split("\n")),))
            ok = False
    if not ok:
        log.error("One or more confetti failed")
    return ok



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.parse
import re

from datetime import datetime

from . import BaseTracker
from . import WithLogin
from . import WithCaptcha
from . import WithCheckTime
from . import WithFetchByTorrentId


# =====
class Plugin(BaseTracker, WithLogin, WithCaptcha, WithCheckTime, WithFetchByTorrentId):
    PLUGIN_NAME = "pornolab.net"

    _SITE_VERSION = 2
    _SITE_ENCODING = "cp1251"

    _SITE_FINGERPRINT_URL = "https://pornolab.net/forum/index.php"
    _SITE_FINGERPRINT_TEXT = "title=\"Поиск на Pornolab.net\" href=\"//static.pornolab.net/opensearch.xml\""

    _COMMENT_REGEXP = re.compile(r"https?://pornolab\.net/forum/viewtopic\.php\?t=(?P<torrent_id>\d+)")

    _TIMEZONE_URL = "https://pornolab.net/forum/index.php"
    _TIMEZONE_REGEXP = re.compile(r"<p>Часовой пояс: <span class='tz_time'>(?P<timezone>GMT [+-] \d{1,2})</span></p>")
    _TIMEZONE_PREFIX = "Etc/"

    _DOWNLOAD_URL = "https://pornolab.net/forum/dl.php?t={torrent_id}"
    _DOWNLOAD_PAYLOAD = b""

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    def fetch_time(self, torrent):
        self._assert_match(torrent)
        page = self._decode(self._read_url(torrent.get_comment()))
        date_match = re.search(r"<span title=\"Зарегистрирован\">\[ (\d\d-([а-яА-Я]{3})-\d\d \d\d:\d\d) \]</span>", page)
        self._assert_logic(date_match is not None, "Upload date not found")
        date = date_match.group(1)
        date_month = date_match.group(2)

        date = date.replace(date_month, {
            month: str(number) for (number, month) in enumerate(
                ("Янв", "Фев", "Мар", "Апр", "Май", "Июн",
                 "Июл", "Авг", "Сен", "Окт", "Ноя", "Дек"), 1
            )
        }[date_month])  # TODO: Send shitbeam to datetime authors
        date += " " + datetime.now(self._tzinfo).strftime("%z")

        upload_time = int(datetime.strptime(date, "%d-%m-%y %H:%M %z").strftime("%s"))
        return upload_time

    def login(self):
        self._assert_required_user_passwd()

        post = {
            "login_username": self._encode(self._user),
            "login_password": self._encode(self._passwd),
            "login":          b"\xc2\xf5\xee\xe4",
        }
        page = self._read_login(post)

        cap_static_regexp = re.compile(r"\"(https?://static\.pornolab\.net/captcha/[^\"]+)\"")
        cap_static_match = cap_static_regexp.search(page)
        if cap_static_match is not None:
            cap_sid_match = re.search(r"name=\"cap_sid\" value=\"([a-zA-Z0-9]+)\"", page)
            cap_code_match = re.search(r"name=\"(cap_code_[a-zA-Z0-9]+)\"", page)
            self._assert_auth(cap_sid_match is not None, "Unknown cap_sid")
            self._assert_auth(cap_code_match is not None, "Unknown cap_code")

            post[cap_code_match.group(1)] = self._captcha_decoder(cap_static_match.group(1))
            post["cap_sid"] = cap_sid_match.group(1)
            page = self._read_login(post)
            self._assert_auth(cap_static_regexp.search(page) is None, "Invalid user, password or captcha")

    def _read_login(self, post):
        return self._decode(self._read_url(
            url="https://pornolab.net/forum/login.php",
            data=self._encode(urllib.parse.urlencode(post)),
        ))



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from ...optconf import Option

from . import BaseTracker
from . import WithCheckHash
from . import WithFetchByTorrentId


# =====
class Plugin(BaseTracker, WithCheckHash, WithFetchByTorrentId):
    PLUGIN_NAME = "rutor.info"

    _SITE_VERSION = 5
    _SITE_ENCODING = "utf-8"

    _SITE_FINGERPRINT_URL = "http://rutor.info"
    _SITE_FINGERPRINT_TEXT = "<a href=\"/\"><img src=\"http://s.rutor.info/logo.jpg\" alt=\"rutor.info logo\" /></a>"

    _COMMENT_REGEXP = re.compile(r"^http://rutor\.(info|org|is)/torrent/(?P<torrent_id>\d+)$")

    _TORRENT_HASH_URL = "http://rutor.info/torrent/{torrent_id}"
    _TORRENT_HASH_REGEXP = re.compile(r"<div id=\"download\">\s+<a href=\"magnet:"
                                      r"\?xt=urn:btih:(?P<torrent_hash>[a-fA-F0-9]{40})")

    _DOWNLOAD_URL = "http://rutor.info/download/{torrent_id}"

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=False)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options({
            "user_agent": Option(default="Googlebot/2.1", help="User-agent for site"),
        })



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import socket
import xmlrpc.client
import argparse


# =====
def manage_trackers(client_url, to_enable, to_disable):
    server = xmlrpc.client.ServerProxy(client_url)

    multicall = xmlrpc.client.MultiCall(server)
    hashes = server.download_list()
    for t_hash in hashes:
        multicall.t.multicall(t_hash, "", "t.is_enabled=", "t.get_url=")
    trackers = list(multicall())

    actions = dict.fromkeys(set(to_enable or ()), 1)
    actions.update(dict.fromkeys(set(to_disable or ()), 0))

    multicall = xmlrpc.client.MultiCall(server)
    for (count, t_hash) in enumerate(hashes):
        for (index, (is_enabled, url)) in enumerate(trackers[count]):
            for (pattern, action) in actions.items():
                if pattern in url and action != is_enabled:
                    multicall.t.set_enabled(t_hash, index, action)
                    print(url, pattern, action)
                    continue
    multicall()


# ===== Main =====
def main():
    args_parser = argparse.ArgumentParser(description="Manage trackers (rtorrent only)")
    args_parser.add_argument("--enable", nargs="+", metavar="<pattern>")
    args_parser.add_argument("--disable", nargs="+", metavar="<pattern>")
    args_parser.add_argument("-t", "--timeout", default=5.0, type=float, metavar="<seconds>")
    args_parser.add_argument("--client-url", default="http://localhost/RPC2", metavar="<url>")

    options = args_parser.parse_args(sys.argv[1:])
    socket.setdefaulttimeout(options.timeout)
    manage_trackers(
        client_url=options.client_url,
        to_enable=options.enable,
        to_disable=options.disable,
    )


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import http.cookiejar
import re

from . import BaseTracker
from . import WithLogin
from . import WithCheckHash
from . import WithFetchCustom


# =====
class Plugin(BaseTracker, WithLogin, WithCheckHash, WithFetchCustom):
    PLUGIN_NAME = "pravtor.ru"

    _SITE_VERSION = 0
    _SITE_ENCODING = "cp1251"

    _SITE_FINGERPRINT_URL = "http://pravtor.ru"
    _SITE_FINGERPRINT_TEXT = "<img src=\"/images/pravtor_beta1.png\""

    _COMMENT_REGEXP = re.compile(r"http://pravtor\.(ru|spb\.ru)/viewtopic\.php\?p=(?P<torrent_id>\d+)")

    _TORRENT_HASH_URL = "http://pravtor.ru/viewtopic.php?p={torrent_id}"
    _TORRENT_HASH_REGEXP = re.compile(r"<span id=\"tor-hash\">(?P<torrent_hash>[a-zA-Z0-9]+)</span>")

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)
        torrent_id = self._COMMENT_REGEXP.match(torrent.get_comment()).group("torrent_id")

        page = self._decode(self._read_url(torrent.get_comment()))
        dl_id_match = re.search(r"<a href=\"download.php\?id=(\d+)\" class=\"(leech|seed|gen)med\">", page)
        self._assert_logic(dl_id_match is not None, "Torrent-ID not found")
        dl_id = int(dl_id_match.group(1))

        cookie = http.cookiejar.Cookie(
            version=0,
            name="bb_dl",
            value=torrent_id,
            port=None,
            port_specified=False,
            domain="",
            domain_specified=False,
            domain_initial_dot=False,
            path="/",
            path_specified=True,
            secure=False,
            expires=None,
            discard=True,
            comment=None,
            comment_url=None,
            rest={"HttpOnly": None},
            rfc2109=False,
        )
        self._cookie_jar.set_cookie(cookie)

        data = self._read_url(
            url="http://pravtor.ru/download.php?id={}".format(dl_id),
            data=b"",
            headers={
                "Referer": "http://pravtor.ru/viewtopic.php?t={}".format(torrent_id),
                "Origin":  "http://pravtor.ru",
            }
        )
        self._assert_valid_data(data)
        return data

    def login(self):
        self._login_using_post(
            url="http://pravtor.ru/login.php",
            post={
                "login_username": self._encode(self._user),
                "login_password": self._encode(self._passwd),
                "login":          b"\xc2\xf5\xee\xe4",
            },
            ok_text="<b class=\"med\">{}</b></a>&nbsp; [ <a href=\"./login.php?logout=1\"".format(self._user),
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import json
import shutil
import contextlib
import traceback
import threading
import operator
import argparse

from ..plugins.trackers import TrackerError
from ..plugins.trackers import WithCheckHash
from ..plugins.trackers import WithCheckScrape
from ..plugins.trackers import WithCheckTime

from ..helpers import tcollection
from ..helpers import surprise

from .. import tfile
from .. import fmt
from .. import tools

from . import init
from . import get_configured_log
from . import get_configured_client
from . import get_configured_trackers
from . import get_configured_confetti


# =====
class OpContext:
    def __init__(self, torrent, tracker):
        self.torrent = torrent
        self.tracker = tracker

        self._status = None
        self._attrs = {}

    def done_not_in_client(self):
        self._status = "not_in_client"

    def done_affected(self, diff):
        self._status = "affected"
        self._attrs = {"diff": diff}

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc, tb):
        if isinstance(exc, TrackerError):
            self._status = "tracker_error"
            self._attrs = {
                "err_name": type(exc).__name__,
                "err_msg": str(exc),
            }
        elif exc is not None:
            self._status = "unhandled_error"
            self._attrs = {"tb_lines": "".join(traceback.format_exception(exc_type, exc, tb)).strip().split("\n")}
        if self._status is None:
            self._status = "passed"


class Feeder:  # pylint: disable=too-many-instance-attributes
    def __init__(self, trackers, torrents, show_unknown, show_passed, show_diff, log_stdout, log_stderr):
        self._trackers = trackers
        self._torrents = torrents
        self._show_unknown = show_unknown
        self._show_passed = show_passed
        self._show_diff = show_diff
        self._log_stdout = log_stdout
        self._log_stderr = log_stderr

        self._current_count = 0
        self._current_file_name = None
        self._current_torrent = None
        self._current_tracker = None

        self._fan = fmt.make_fan()
        self._fan_thread = None
        self._stop_fan = threading.Event()

        self._status_mapping = {
            "invalid":         self._done_invalid,
            "not_in_client":   self._done_not_in_client,
            "unknown":         self._done_unknown,
            "passed":          self._done_passed,
            "affected":        self._done_affected,
            "tracker_error":   self._done_tracker_error,
            "unhandled_error": self._done_unhandled_error,
        }

        self._results = {status: {} for status in self._status_mapping}

    def get_ops(self):
        for (self._current_count, (self._current_file_name, self._current_torrent)) in enumerate(
            sorted(self._torrents.items(), key=operator.itemgetter(0))
        ):
            self._current_tracker = None

            if self._current_torrent is None:
                self._done("invalid", {})
                continue

            self._current_tracker = self._select_tracker()
            if self._current_tracker is None:
                self._done("unknown", {})
                continue

            op = OpContext(self._current_torrent, self._current_tracker)
            self._start_op()
            yield op
            self._stop_op()
            self._done(op._status, op._attrs)  # pylint: disable=protected-access

        self._log_stdout.finish()

    def get_results(self):
        return self._results

    def _select_tracker(self):
        for tracker in self._trackers:
            if tracker.is_matched_for(self._current_torrent):
                return tracker
        return None

    def _done(self, status, attrs):
        self._status_mapping[status](attrs)
        result = {
            "torrent": self._current_torrent,
            "tracker": self._current_tracker,
        }
        result.update(attrs)
        self._results[status][self._current_file_name] = result

    def _done_invalid(self, _):
        self._log_stdout.print(*self._format_fail("red", "!", "INVALID_TORRENT"))

    def _done_unknown(self, _):
        one_line = (not self._show_unknown and self._log_stdout.isatty())
        self._log_stdout.print(*self._format_fail("yellow", " ", "UNKNOWN"), one_line=one_line)

    def _done_passed(self, _):
        one_line = (not self._show_passed and self._log_stdout.isatty())
        self._log_stdout.print(*self._format_status("blue", " "), one_line=one_line)

    def _done_affected(self, attrs):
        self._log_stdout.print(*self._format_status("green", "+"))
        if self._show_diff:
            self._log_stdout.print(*fmt.format_torrents_diff(attrs["diff"], "\t"))

    def _done_not_in_client(self, _):
        self._log_stdout.print(*self._format_fail("red", "!", "NOT_IN_CLIENT"))

    def _done_tracker_error(self, attrs):
        (line, placeholders) = self._format_status("red", "-")
        line += " :: {red}%s({reset}%s{red}){reset}"
        placeholders += (attrs["err_name"], attrs["err_msg"])
        self._log_stdout.print(line, placeholders)

    def _done_unhandled_error(self, attrs):
        self._log_stdout.print(*self._format_status("red", "-"))
        self._log_stdout.print("%s", ("\n".join("\t" + row for row in attrs["tb_lines"]),))

    def _start_op(self):
        if self._log_stdout.isatty():
            def loop():
                while not self._stop_fan.wait(timeout=0.1):
                    self._log_stdout.print(*self._format_status("magenta", next(self._fan)), one_line=True)
            self._fan_thread = threading.Thread(target=loop, daemon=True)
            self._fan_thread.start()
        else:
            self._log_stdout.print(*self._format_status("magenta", " "))

    def _stop_op(self):
        if self._fan_thread is not None:
            self._stop_fan.set()
            self._fan_thread.join()
            self._stop_fan.clear()

    def _format_fail(self, color, sign, error):
        return (
            "[{" + color + "}%s{reset}] %s {" + color + "}%s {cyan}%s{reset}",
            (sign, self._format_progress(), error, self._current_file_name),
        )

    def _format_status(self, color, sign):
        return (
            "[{" + color + "}%s{reset}] %s {" + color + "}%s {cyan}%s{reset} -- %s",
            (
                sign, self._format_progress(), self._current_tracker.PLUGIN_NAME,
                self._current_file_name, (self._current_torrent.get_comment() or ""),
            ),
        )

    def _format_progress(self):
        return fmt.format_progress(self._current_count + 1, len(self._torrents))


def backup_torrent(torrent, backup_dir_path, backup_suffix):
    backup_suffix = fmt.format_now(backup_suffix)
    backup_file_path = os.path.join(backup_dir_path, os.path.basename(torrent.get_path()) + backup_suffix)
    shutil.copyfile(torrent.get_path(), backup_file_path)


@contextlib.contextmanager
def client_hooks(client, torrent, to_save_customs, to_set_customs):
    if client is not None:
        prefix = client.get_data_prefix(torrent)
        if len(to_save_customs) != 0:
            customs = client.get_customs(torrent, to_save_customs)
        else:
            customs = {}
        meta_file_path = tools.make_sub_name(torrent.get_path(), ".", ".meta")
        with open(meta_file_path, "w") as meta_file:
            meta_file.write(json.dumps({
                "prefix": prefix,
                "customs": customs,
            }))
        client.remove_torrent(torrent)

    yield  # Torrent-object was changed

    if client is not None:
        client.load_torrent(torrent, prefix)
        customs.update({
            key: fmt.format_now(value)
            for (key, value) in to_set_customs.items()
        })
        if len(customs) != 0:
            client.set_customs(torrent, customs)
        os.remove(meta_file_path)


def update_torrent(client, torrent, new_data, to_save_customs, to_set_customs):
    data_file_path = tools.make_sub_name(torrent.get_path(), ".", ".newdata")
    with open(data_file_path, "wb") as data_file:
        data_file.write(new_data)
    with client_hooks(client, torrent, to_save_customs, to_set_customs):
        os.replace(data_file_path, torrent.get_path())
        torrent.load_from_data(new_data, torrent.get_path())


class TorrentTimeInfo:
    def __init__(self, torrent):
        self._torrent = torrent
        self._time_file_path = tools.make_sub_name(self._torrent.get_path(), ".", ".time")

    def check_and_fill(self):
        if not os.path.exists(self._time_file_path):
            self.write(int(os.stat(self._torrent.get_path()).st_mtime))
        return self

    def read(self):
        with open(self._time_file_path) as time_file:
            return int(time_file.read().strip())

    def write(self, value):
        with open(self._time_file_path, "w") as time_file:
            time_file.write(str(int(value)))


def update(  # pylint: disable=too-many-branches,too-many-locals
    feeder,
    client,
    backup_dir_path,
    backup_suffix,
    to_save_customs,
    to_set_customs,
    noop,
    test_mode,
):
    hashes = (client.get_hashes() if client is not None else [])
    for op in feeder.get_ops():
        try:
            with op:
                if client is not None and op.torrent.get_hash() not in hashes:
                    op.done_not_in_client()
                    continue

                tracker_bases = op.tracker.get_bases()

                if WithCheckHash in tracker_bases:
                    need_update = (op.tracker.fetch_hash(op.torrent) != op.torrent.get_hash())

                elif WithCheckScrape in tracker_bases:
                    need_update = (not op.tracker.is_registered(op.torrent))

                elif WithCheckTime in tracker_bases:
                    time_info = TorrentTimeInfo(op.torrent).check_and_fill()
                    tracker_time = op.tracker.fetch_time(op.torrent)
                    need_update = (tracker_time > time_info.read())

                else:
                    RuntimeError("Invalid tracker {}: missing method of check".format(op.tracker))

                if need_update or test_mode:
                    tmp_torrent = tfile.Torrent(data=op.tracker.fetch_new_data(op.torrent))
                    if op.torrent.get_hash() != tmp_torrent.get_hash() or test_mode:
                        diff = tfile.get_difference(op.torrent, tmp_torrent)
                        if not noop:
                            if backup_dir_path is not None:
                                backup_torrent(op.torrent, backup_dir_path, backup_suffix)
                            update_torrent(client, op.torrent, tmp_torrent.get_data(), to_save_customs, to_set_customs)
                            if WithCheckTime in tracker_bases:
                                time_info.write(tracker_time)
                        op.done_affected(diff)
                    elif WithCheckTime in tracker_bases and not noop:
                        time_info.write(tracker_time)
        except Exception:
            pass


def print_results(results, log):
    for (msg, field) in (
        ("Updated:          %d", "affected"),
        ("Passed:           %d", "passed"),
        ("Not in client:    %d", "not_in_client"),
        ("Unknown:          %d", "unknown"),
        ("Invalid torrents: %d", "invalid"),
        ("Tracker errors:   %d", "tracker_error"),
        ("Unhandled errors: %d", "unhandled_error"),
    ):
        log.info(msg, (len(results[field]),))


# ===== Main =====
def main():
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emupdate",
        description="Update torrent files from popular trackers",
        parents=[parent_parser],
    )
    args_parser.add_argument("-f", "--name-filter", default=None, metavar="<wildcard_pattern>")
    args_parser.add_argument("-y", "--only-trackers", default=[], nargs="+", metavar="<tracker>")
    args_parser.add_argument("-x", "--exclude-trackers", default=[], nargs="+", metavar="<tracker>")
    args_parser.add_argument("--noop", action="store_true")
    args_parser.add_argument("--mute", action="store_true")
    args_parser.add_argument("--fail-on-captcha", action="store_true")
    args_parser.add_argument("--test-mode", action="store_true", help=argparse.SUPPRESS)
    options = args_parser.parse_args(argv[1:])

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, False, sys.stderr) as log_stderr:

            client = get_configured_client(
                config=config,
                required=False,
                with_customs=(len(config.emupdate.save_customs) or len(config.emupdate.set_customs)),
                log=log_stderr,
            )

            def read_captcha(url):
                if options.fail_on_captcha:
                    raise RuntimeError("Required decoding of captcha but '--fail-on-captcha' specified")
                else:
                    log_stderr.info("{yellow}Enter the captcha{reset} from [{blue}%s{reset}]: ", (url,), no_nl=True)
                    return input()

            trackers = get_configured_trackers(
                config=config,
                captcha_decoder=read_captcha,
                only=options.only_trackers,
                exclude=options.exclude_trackers,
                log=log_stderr,
            )

            if not options.mute:
                confetti = get_configured_confetti(
                    config=config,
                    log=log_stderr,
                )

            torrents = tcollection.load_from_dir(
                path=config.core.torrents_dir,
                name_filter=(options.name_filter or config.emupdate.name_filter),
                log=log_stderr,
            )

            feeder = Feeder(
                trackers=trackers,
                torrents=torrents,
                show_unknown=config.emupdate.show_unknown,
                show_passed=config.emupdate.show_passed,
                show_diff=config.emupdate.show_diff,
                log_stdout=log_stdout,
                log_stderr=log_stderr,
            )

            update(
                feeder=feeder,
                client=client,
                backup_dir_path=config.emupdate.backup_dir,
                backup_suffix=config.emupdate.backup_suffix,
                to_save_customs=config.emupdate.save_customs,
                to_set_customs=config.emupdate.set_customs,
                noop=options.noop,
                test_mode=options.test_mode,
            )

            results = feeder.get_results()
            print_results(
                results=results,
                log=log_stderr,
            )
            if not options.mute:
                if len(results["affected"]) != 0:
                    if not surprise.deploy_surprise(
                        source="emupdate",
                        results=results,
                        confetti=confetti,
                        log=log_stderr,
                    ):
                        raise SystemExit(1)


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from datetime import datetime

from . import BaseTracker
from . import WithLogin
from . import WithCaptcha
from . import WithCheckTime
from . import WithFetchByDownloadId


# =====
class Plugin(BaseTracker, WithLogin, WithCaptcha, WithCheckTime, WithFetchByDownloadId):
    PLUGIN_NAME = "trec.to"

    _SITE_VERSION = 1
    _SITE_ENCODING = "utf-8"

    _SITE_FINGERPRINT_URL = "http://trec.to"
    _SITE_FINGERPRINT_TEXT = "var cookieDomain  = \"trec.to\";"

    _COMMENT_REGEXP = re.compile(r"http://(torrent\.rus\.ec|trec\.to)/viewtopic\.php\?p=(?P<torrent_id>\d+)")

    _TIMEZONE_URL = "http://trec.to"
    _TIMEZONE_REGEXP = re.compile(r"<p>Часовой пояс: <span class=\"tz_time\">(?P<timezone>GMT [+-] \d{1,2})</span></p>")
    _TIMEZONE_PREFIX = "Etc/"

    _DOWNLOAD_ID_URL = "http://trec.to/viewtopic.php?p={torrent_id}"
    _DOWNLOAD_ID_REGEXP = re.compile(r"<a href=\"download\.php\?id=(?P<download_id>\d+)\" class=\"(leech|seed|gen)med\">")
    _DOWNLOAD_URL = "http://trec.to/download.php?id={download_id}"

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    def fetch_time(self, torrent):
        self._assert_match(torrent)
        torrent_id = self._COMMENT_REGEXP.match(torrent.get_comment()).group("torrent_id")
        page = self._decode(self._read_url("http://trec.to/viewtopic.php?p={}".format(torrent_id)))

        date_match = re.search(r"<td width=\"70%\">\s*Зарегистрирован &nbsp;\s*\[ <span title=\"\">"
                               r"(\d\d-\d\d-\d\d\d\d \d\d:\d\d)</span> ]\s*</td>", page)
        self._assert_logic(date_match is not None, "Upload date not found")
        date = date_match.group(1)
        date += " " + datetime.now(self._tzinfo).strftime("%z")
        upload_time = int(datetime.strptime(date, "%d-%m-%Y %H:%M %z").strftime("%s"))
        return upload_time

    def login(self):
        self._login_using_post(
            url="http://trec.to/login.php",
            post={
                "login_username": self._encode(self._user),
                "login_password": self._encode(self._passwd),
                "login":          self._encode("Вход"),
            },
            ok_text="<a href=\"./login.php?logout=1\" onclick=\"return confirm",
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import argparse

from ..helpers import tcollection

from .. import tfile
from .. import fmt

from . import init
from . import get_configured_log
from . import get_configured_client


# ===== Main =====
def main():
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emdiff",
        description="Show a difference between two torrent files",
        parents=[parent_parser],
    )
    args_parser.add_argument("-v", "--verbose", action="store_true")
    args_parser.add_argument("torrents", type=str, nargs=2, metavar="<path/hash>")
    options = args_parser.parse_args(argv[1:])

    torrents = tcollection.find(config.core.torrents_dir, options.torrents, True)

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, (not options.verbose), sys.stderr) as log_stderr:
            client = None
            lists = []
            for item in torrents:
                if isinstance(item, tfile.Torrent):
                    lists.append(item.get_files())
                else:  # Hash
                    if client is None:
                        client = get_configured_client(
                            config=config,
                            required=True,
                            with_customs=False,
                            log=log_stderr,
                        )
                    lists.append(client.get_files(item))
            diff = tfile.get_difference(*lists)  # pylint: disable=no-value-for-parameter
            log_stdout.print(*fmt.format_torrents_diff(diff, " "))


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import urllib.request
import urllib.parse
import http.client
import socket

from ..thirdparty import socks


# =====
SCHEME_TO_TYPE = {
    "socks4": socks.PROXY_TYPE_SOCKS4,
    "socks5": socks.PROXY_TYPE_SOCKS5,
}

SOCKS_PORT = 1080


# =====
class SocksHandler(urllib.request.HTTPHandler, urllib.request.HTTPSHandler):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        super().__init__(debuglevel=kwargs.pop("debuglevel", 0))

    def http_open(self, req):
        def build(host, port=None, strict=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):  # pylint: disable=protected-access
            connection = _SocksConnection(host, port=port, strict=strict, timeout=timeout, **self._kwargs)
            connection.make_proxy_args(*self._args, **self._kwargs)
            return connection
        return self.do_open(build, req)

    def https_open(self, req):
        def build(host, port=None, strict=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):  # pylint: disable=protected-access
            connection = _SocksSecureConnection(host, port=port, strict=strict, timeout=timeout, **self._kwargs)
            connection.make_proxy_args(*self._args, **self._kwargs)
            return connection
        return self.do_open(build, req)


# =====
class _SocksConnection(http.client.HTTPConnection):
    def __init__(self, *args, **kwargs):
        kwargs.pop("strict", None)  # XXX: Fix for "TypeError: __init__() got an unexpected keyword argument 'strict'"
        kwargs.pop("proxy_url", None)  # XXX: Fix for "TypeError: __init__() got an unexpected keyword argument 'proxy_url'"
        super().__init__(*args, **kwargs)
        self._proxy_args = None

    # XXX: because proxy args/kwargs break super
    def make_proxy_args(
        self,
        proxy_url=None,
        proxy_type=None,
        proxy_host=None,
        proxy_port=None,
        proxy_user=None,
        proxy_passwd=None,
        rdns=True
    ):
        if proxy_url is not None:
            parsed = urllib.parse.urlparse(proxy_url)
            scheme = parsed.scheme
            proxy_user = parsed.username
            proxy_passwd = parsed.password
            proxy_host = parsed.hostname
            proxy_port = (parsed.port or SOCKS_PORT)
            proxy_type = SCHEME_TO_TYPE.get((scheme or "").lower())
            if proxy_type is None:
                raise RuntimeError("Invalid SOCKS protocol: {}".format(scheme))

        self._proxy_args = (proxy_type, proxy_host, proxy_port, rdns, proxy_user, proxy_passwd)

    def connect(self):
        if self._proxy_args is None:
            raise RuntimeError("Proxy args weren't initialized")
        self.sock = socks.socksocket()
        self.sock.setproxy(*self._proxy_args)
        if self.timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:  # pylint: disable=protected-access
            self.sock.settimeout(self.timeout)
        self.sock.connect((self.host, self.port))


# =====
class _SocksSecureConnection(http.client.HTTPSConnection, _SocksConnection):
    def __init__(self, *args, **kwargs):
        kwargs.pop("strict", None)  # XXX: Fix for "TypeError: __init__() got an unexpected keyword argument 'strict'"
        kwargs.pop("proxy_url", None)  # XXX: Fix for "TypeError: __init__() got an unexpected keyword argument 'proxy_url'"
        super().__init__(*args, **kwargs)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re
import operator

from datetime import datetime

from dateutil.relativedelta import relativedelta

from ... import tfile

from . import BaseTracker
from . import WithLogin
from . import WithCheckTime
from . import WithFetchCustom


# =====
class Plugin(BaseTracker, WithLogin, WithCheckTime, WithFetchCustom):
    PLUGIN_NAME = "tr.anidub.com"

    _SITE_VERSION = 0
    _SITE_ENCODING = "utf-8"

    _SITE_FINGERPRINT_URL = "http://tr.anidub.com"
    _SITE_FINGERPRINT_TEXT = "href=\"http://tr.anidub.com/engine/opensearch.php\" title=\"AniDUB Tracker\""

    _COMMENT_REGEXP = re.compile(r"http://tr\.anidub\.com/\?newsid=(?P<torrent_id>\d+)")

    _TIMEZONE_STATIC = "Etc/GMT+4"

    # ===

    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls):
        return cls._get_merged_options()

    def fetch_time(self, torrent):
        self._assert_match(torrent)
        page = self._decode(self._read_url(torrent.get_comment()))
        date_match = re.search(r"<li><b>Дата:</b> ([^,\s]+, \d\d:\d\d)</li>", page)
        self._assert_logic(date_match is not None, "Upload date not found")
        date = date_match.group(1)

        now = datetime.now(self._tzinfo)
        day_template = "{date.day:02d}-{date.month:02d}-{date.year}"
        if "Сегодня" in date:
            date = date.replace("Сегодня", day_template.format(date=now))
        if "Вчера" in date:
            yesterday = now - relativedelta(days=1)
            date = date.replace("Вчера", day_template.format(date=yesterday))
        date += " " + datetime.now(self._tzinfo).strftime("%z")

        upload_time = int(datetime.strptime(date, "%d-%m-%Y, %H:%M %z").strftime("%s"))
        return upload_time

    def fetch_new_data(self, torrent):
        self._assert_match(torrent)
        page = self._decode(self._read_url(torrent.get_comment()))
        downloads = set(map(int, re.findall(r"<a href=\"/engine/download.php\?id=(\d+)\" class=\" \">", page)))
        candidates = {}
        for download_id in downloads:
            data = self._read_url(
                url="http://tr.anidub.com/engine/download.php?id={}".format(download_id),
                headers={"Referer": torrent.get_comment()},
            )
            self._assert_valid_data(data)
            candidate = tfile.Torrent(data=data)

            name = candidate.get_name()
            candidates.setdefault(name, [])
            candidates[name].append((candidate, str(download_id)))

        name = torrent.get_name()
        self._assert_logic(name in candidates, "Can't find torrent named '{}' in downloads".format(name))
        self._assert_logic(len(candidates[name]) == 1, "Too many variants to download: {}".format(
                                                       ", ".join(map(operator.itemgetter(1), candidates[name]))))
        return candidates[name][0][0].get_data()

    def login(self):
        self._login_using_post(
            url="http://tr.anidub.com/",
            post={
                "login_name":      self._encode(self._user),
                "login_password":  self._encode(self._passwd),
                "login":           b"submit",
            },
            ok_text="<li><a href=\"http://tr.anidub.com/user/{}/\">Мой профиль</a></li>".format(self._user)
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import re

from colorama import Fore
from colorama import Style


# =====
_COLORS = {
    "red":     Style.BRIGHT + Fore.RED,
    "green":   Style.BRIGHT + Fore.GREEN,
    "yellow":  Style.BRIGHT + Fore.YELLOW,
    "cyan":    Style.BRIGHT + Fore.CYAN,
    "magenta": Style.BRIGHT + Fore.MAGENTA,
    "blue":    Style.BRIGHT + Fore.BLUE,
    "reset":   Fore.RESET,
}

_NO_COLORS = dict.fromkeys(list(_COLORS), "")


# =====
class Log:
    def __init__(self, use_colors=True, force_colors=False, quiet=False, output=sys.stdout):
        self._quiet = quiet
        self._output = output
        self._fill = 0
        self._colored = (use_colors and (self.isatty() or force_colors))
        self._ansi_regexp = re.compile(r"(\x1b[^m]*m)")

    def isatty(self):
        return self._output.isatty()

    def info(self, text, *args, **kwargs):
        self.print("# {green}I{reset}: " + text, *args, **kwargs)

    def error(self, text, *args, **kwargs):
        self.print("# {red}E{reset}: " + text, *args, **kwargs)

    def print(self, text="", placeholders=(), one_line=False, no_nl=False):
        if not self._quiet:
            self.finish()

            rendered = self._format_text(text, placeholders, self._colored)
            view_len = len(self._format_text(text, placeholders, False) if self._colored else rendered)

            cut = 0
            if one_line and self.isatty():
                max_len = self._get_term_width()
                if max_len is not None and view_len > max_len:
                    cut = view_len - max_len
                    rendered = self._cut_line(rendered, cut)

            self._output.write(rendered)
            if one_line and self.isatty():
                self._output.write("\r")
                self._fill = view_len - cut
            elif not no_nl:
                self._output.write("\n")
                self._fill = 0
            self._output.flush()

    def finish(self):
        if self._fill:
            self._output.write((" " * self._fill) + "\r")
            self._fill = 0
            self._output.flush()

    def _get_term_width(self):
        try:
            return int(os.environ["COLUMNS"])
        except (KeyError, ValueError):
            try:
                return os.get_terminal_size(self._output.fileno()).columns
            except OSError:
                return None

    def _format_text(self, text, placeholders, colored):
        text = text.format(**(_COLORS if colored else _NO_COLORS))
        text = text % placeholders
        return text

    def _cut_line(self, text, cut):
        parts = self._ansi_regexp.split(text)
        for index in reversed(range(len(parts))):
            if cut <= 0:
                break
            if parts[index].startswith("\x1b"):
                parts[index] = ""
            elif cut <= len(parts[index]):
                parts[index] = parts[index][:-cut]
                break
            else:
                cut -= len(parts[index])
                parts[index] = ""
        if self._colored:
            parts.append(Fore.RESET)
        return "".join(parts)
